from __future__ import annotations
from typing import TypeVar
import astropy.units as u
import named_arrays as na

__all__ = [
    "RandomLowT",
    "RandomHighT",
    "uniform",
    "normal",
    "poisson",
]


RandomLowT = TypeVar("RandomLowT", bound="float | complex | u.Quantity | na.AbstractArray")
RandomHighT = TypeVar("RandomHighT", bound="float | complex | u.Quantity | na.AbstractArray")
RandomCenterT = TypeVar("RandomCenterT", bound="float | complex | u.Quantity | na.AbstractArray")
RandomWidthT = TypeVar("RandomWidthT", bound="float | complex | u.Quantity | na.AbstractArray")


def uniform(
        low: RandomLowT,
        high: RandomHighT,
        shape_random: dict[str, int] = None,
        seed: None | int = None
) -> RandomLowT | RandomHighT:
    """
    Draw samples from a uniform distribution

    Parameters
    ----------
    low
        Lower boundary of the output
    high
        Upper boundary of the output
    shape_random
        Additional dimensions to be broadcast against :attr:`start` and :attr:`stop`
    seed
        Seed for the random number generator, can be provided for repeatability

    See Also
    --------
    :func:`numpy.random.uniform` : Equivalent numpy function

    :class:`named_arrays.ScalarUniformRandomSample` : Implicit array version of this function for scalars

    :class:`named_arrays.UncertainScalarUniformRandomSample` : Implicit array version of this function for uncertain
        scalars
    """

    return na._named_array_function(
        func=uniform,
        low=na.as_named_array(low),
        high=na.as_named_array(high),
        shape_random=shape_random,
        seed=seed,
    )


def normal(
        loc: RandomCenterT,
        scale: RandomWidthT,
        shape_random: dict[str, int] = None,
        seed: None | int = None
) -> RandomCenterT | RandomWidthT:
    """
    Draw samples from a normal distribution

    Parameters
    ----------
    loc
        The center of the distribution.
    scale
        The width of the distribution.
    shape_random
        Additional dimensions to be broadcast against :attr:`center` and :attr:`width`
    seed
        Seed for the random number generator, can be provided for repeatability.

    See Also
    --------
    :func:`numpy.random.normal` : Equivalent numpy function

    :class:`named_arrays.ScalarNormalRandomSample` : Implicit array version of this function for scalars

    :class:`named_arrays.UncertainScalarNormalRandomSample` : Implicit array version of this function for uncertain
        scalars
    """
    return na._named_array_function(
        func=normal,
        loc=na.as_named_array(loc),
        scale=na.as_named_array(scale),
        shape_random=shape_random,
        seed=seed,
    )


def poisson(
        lam: RandomCenterT,
        shape_random: dict[str, int] = None,
        seed: None | int = None
) -> RandomCenterT:
    """
    Draw samples from a Poisson distribution

    Parameters
    ----------
    lam
        Expected number of events
    shape_random
        Additional dimensions to be broadcast against :attr:`center` and :attr:`width`
    seed
        Seed for the random number generator, can be provided for repeatability.

    See Also
    --------
    :func:`numpy.random.poisson` : Equivalent numpy function

    :class:`named_arrays.ScalarPoissonlRandomSample` : Implicit array version of this function for scalars

    :class:`named_arrays.UncertainScalarPoissonRandomSample` : Implicit array version of this function for uncertain
        scalars
    """
    return na._named_array_function(
        func=poisson,
        lam=na.as_named_array(lam),
        shape_random=shape_random,
        seed=seed,
    )


from __future__ import annotations
import matplotlib.axes
import matplotlib.artist
import matplotlib.pyplot as plt
import numpy.typing as npt
import named_arrays as na

__all__ = [
    "subplots",
    "plot",
]


def subplots(
        axis_rows: None | str = None,
        ncols: int = 1,
        axis_cols: None | str = None,
        nrows: int = 1,
        *,
        sharex: bool | str = False,
        sharey: bool | str = False,
        squeeze: bool = True,
        **kwargs,
) -> tuple[plt.Figure, na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]]]:
    """
    A thin wrapper around :func:`matplotlib.pyplot.subplots()` which allows for
    providing axis names to the rows and columns.

    Parameters
    ----------
    axis_rows
        Name of the axis representing the rows in the subplot grid.
        If :obj:`None`, the ``squeeze`` argument must be :class:`True`.
    nrows
        Number of rows in the subplot grid
    axis_cols
        Name of the axis representing the columns in the subplot grid
        If :obj:`None`, the ``squeeze`` argument must be :obj:`True`.
    ncols
        Number of columns in the subplot grid
    sharex
        Controls whether all the :class:`matplotlib.axes.Axes` instances share the same horizontal axis properties.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    sharey
        Controls whether all the :class`matplotlib.axes.Axes` instances share the same vertical axis properties.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    squeeze
        If :obj:`True`, :func:`numpy.squeeze` is called on the result, which removes singleton dimensions from the
        array.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    kwargs
        Additional keyword arguments passed to :func:`matplotlib.pyplot.subplots`
    """

    axes = (axis_rows, axis_cols)
    axes = tuple(axis for axis in axes if axis is not None)

    fig, axs = plt.subplots(
        ncols=ncols,
        nrows=nrows,
        sharex=sharex,
        sharey=sharey,
        squeeze=squeeze,
        **kwargs,
    )

    return fig, na.ScalarArray(axs, axes)


def plot(
        *args: na.AbstractScalar,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalar = True,
        **kwargs,
) -> na.ScalarArray[npt.NDArray[None | matplotlib.artist.Artist]]:
    """
    A thin wrapper around :meth:`matplotlib.axes.Axes.plot` for named arrays.

    The main difference of this function from :func:`matplotlib.pyplot.plot` is the addition of the ``axis`` parameter
    indicating along which axis the lines should be connected.

    Parameters
    ----------
    args
        either ``x, y`` or ``y``, same as :meth:`matplotlib.axes.Axes.plot`
    ax
        The instances of :class:`matplotlib.axes.Axes` to use.
        If :obj:`None`, calls :func:`matplotlib.pyplot.gca` to get the current axes.
        If an instance of :class:`named_arrays.ScalarArray`, ``ax.shape`` should be a subset of the broadcasted shape of
        ``*args``.
    axis
        The name of the axis that the plot lines should be connected along.
        If :obj:`None`, the broadcasted shape of ``args`` should have only one element,
        otherwise a :class:`ValueError` is raised.
    where
        A boolean array that selects which elements to plot
    kwargs
        Additional keyword arguments passed to :meth:`matplotlib.axes.Axes.plot`.
        These can be instances of :class:`named_arrays.AbstractArray`.

    Returns
    -------
        An array of artists that were plotted

    Examples
    --------

    Plot a single scalar

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        x = na.linspace(0, 2 * np.pi, axis="x",  num=101)
        y = np.sin(x)

        plt.figure();
        na.plt.plot(x, y);

    Plot an array of scalars

    .. jupyter-execute::

        z = na.linspace(0, np.pi, axis="z", num=5)

        y = np.sin(x - z)

        plt.figure();
        na.plt.plot(x, y, axis="x");

    Plot an uncertain scalar

    .. jupyter-execute::

        ux = na.NormalUncertainScalarArray(x, width=0.2)
        uy = np.sin(ux)

        plt.figure();
        na.plt.plot(x, uy);

    Broadcast an array of scalars against an array of :class:`matplotlib.axes.Axes`

    .. jupyter-execute::

        fig, ax = na.plt.subplots(axis_rows="z", nrows=z.shape["z"], sharex=True)

        na.plt.plot(x, y, ax=ax, axis="x");

    """
    return na._named_array_function(
        plot,
        *args,
        ax=ax,
        axis=axis,
        where=where,
        **kwargs,
    )



from __future__ import annotations
from typing import TypeVar, Generic, Type
import abc
import dataclasses
import named_arrays as na
import numpy as np

__all__ = [
    'AbstractSpectralMatrixArray',
    'SpectralMatrixArray',
]

WavelengthT = TypeVar('WavelengthT', bound=na.AbstractVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSpectralMatrixArray(
    na.AbstractCartesianMatrixArray,
    na.AbstractSpectralVectorArray,
):
    @property
    @abc.abstractmethod
    def wavelength(self) -> na.AbstractVectorArray:
        """
        The `wavelength` component of the matrix.
        """

    @property
    def type_abstract(self) -> Type[AbstractSpectralMatrixArray]:
        return AbstractSpectralMatrixArray

    @property
    def type_explicit(self) -> Type[SpectralMatrixArray]:
        return SpectralMatrixArray

    @property
    def type_vector(self) -> Type[na.SpectralVectorArray]:
        return na.SpectralVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        return np.abs(self.wavelength)

    @property
    def inverse(self) -> na.AbstractMatrixArray:
        return 1 / self


@dataclasses.dataclass(eq=False, repr=False)
class SpectralMatrixArray(
    na.SpectralVectorArray,
    AbstractSpectralMatrixArray,
    na.AbstractExplicitMatrixArray,
    Generic[WavelengthT],
):
    pass



import numpy as np
import pytest
import astropy.units as u
import named_arrays as na
import named_arrays._vectors.tests.test_vectors

__all__ = [
    'AbstractTestAbstractCartesianVectorArray',
    'AbstractTestAbstractExplicitCartesianVectorArray',
    'AbstractTestAbstractImplicitCartesianVectorArray',
    'AbstractTestAbstractCartesianVectorRandomSample',
    'AbstractTestAbstractCartesianVectorUniformRandomSample',
    'AbstractTestAbstractCartesianVectorNormalRandomSample',
    'AbstractTestAbstractParameterizedCartesianVectorArray',
    'AbstractTestAbstractCartesianVectorArrayRange',
    'AbstractTestAbstractCartesianVectorSpace',
    'AbstractTestAbstractCartesianVectorLinearSpace',
    'AbstractTestAbstractCartesianVectorStratifiedRandomSpace',
    'AbstractTestAbstractCartesianVectorLogarithmicSpace',
    'AbstractTestAbstractCartesianVectorGeometricSpace',
]


class AbstractTestAbstractCartesianVectorArray(
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorArray,
):

    def test__mul__(self, array: na.AbstractVectorArray):
        unit = u.mm
        result = array * unit
        for c in array.components:
            assert np.all(result.components[c] == array.components[c] * unit)

    def test__lshift__(self, array: na.AbstractVectorArray):
        unit = u.mm
        result = array << unit
        for c in array.components:
            assert np.all(result.components[c] == array.components[c] << unit)

    def test__truediv__(self, array: na.AbstractVectorArray):
        unit = u.mm
        result = array / unit
        for c in array.components:
            assert np.all(result.components[c] == array.components[c] / unit)

    class TestUfuncUnary(
        named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorArray.TestUfuncUnary,
    ):
        def test_ufunc_unary(
                self,
                ufunc: np.ufunc,
                array: na.AbstractVectorArray,
        ):
            super().test_ufunc_unary(ufunc=ufunc, array=array)

            components = array.components

            kwargs = dict()
            kwargs_components = {c: dict() for c in components}

            try:
                if ufunc in [np.log, np.log2, np.log10, np.sqrt]:
                    kwargs["where"] = array > 0
                elif ufunc in [np.log1p]:
                    kwargs["where"] = array >= -1
                elif ufunc in [np.arcsin, np.arccos, np.arctanh]:
                    kwargs["where"] = (-1 < array) & (array < 1)
                elif ufunc in [np.arccosh]:
                    kwargs["where"] = array >= 1
                elif ufunc in [np.reciprocal]:
                    kwargs["where"] = array != 0
            except u.UnitConversionError:
                pass

            for c in components:
                for k in kwargs:
                    if isinstance(kwargs[k], na.AbstractVectorArray):
                        kwargs_components[c][k] = kwargs[k].components[c]
                    else:
                        kwargs_components[c][k] = kwargs[k]

            result_expected = tuple(array.prototype_vector for _ in range(ufunc.nout))
            try:
                for c in components:
                    result_c = ufunc(components[c], **kwargs_components[c])
                    if ufunc.nout == 1:
                        result_c = (result_c,)
                    for i in range(ufunc.nout):
                        result_expected[i].components[c] = result_c[i]

            except Exception as e:
                with pytest.raises(type(e)):
                    ufunc(array, **kwargs)
                return

            result = ufunc(array, **kwargs)

            if ufunc.nout == 1:
                out = 0 * np.nan_to_num(result)
            else:
                out = tuple(0 * np.nan_to_num(r) for r in result)

            result_out = ufunc(array, out=out, **kwargs)

            if ufunc.nout == 1:
                out = (out,)
                result = (result,)
                result_out = (result_out,)

            for i in range(ufunc.nout):
                assert np.all(result[i] == result_expected[i], **kwargs)
                assert np.all(result[i] == result_out[i], **kwargs)
                assert result_out[i] is out[i]

    class TestUfuncBinary(
        named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorArray.TestUfuncBinary,
    ):

        def test_ufunc_binary(
                self,
                ufunc: np.ufunc,
                array: None | bool | int | float | complex | str | na.AbstractArray | na.AbstractCartesianVectorArray,
                array_2: None | bool | int | float | complex | str | na.AbstractArray | na.AbstractCartesianVectorArray,
        ):
            super().test_ufunc_binary(ufunc, array, array_2)

            if isinstance(array, na.AbstractVectorArray):
                array_normalized = array
                if isinstance(array_2, na.AbstractVectorArray):
                    array_2_normalized = array_2
                else:
                    array_2_normalized = array.type_explicit.from_scalar(array_2, like=array)
            else:
                if isinstance(array_2, na.AbstractVectorArray):
                    array_normalized = array_2.type_explicit.from_scalar(array, like=array_2)
                    array_2_normalized = array_2
                else:
                    raise ValueError("One of the test arrays should be vectors")

            components = array_normalized.components
            components_2 = array_2_normalized.components

            kwargs = dict()
            kwargs_components = {c: dict() for c in components}

            try:
                if ufunc in [np.power, np.float_power]:
                    kwargs["where"] = (array_2 >= 1) & (array >= 0)
                elif ufunc in [np.divide, np.floor_divide, np.remainder, np.fmod, np.divmod]:
                    kwargs["where"] = array_2 != 0
            except (u.UnitConversionError, TypeError):
                pass

            for c in components:
                for k in kwargs:
                    if isinstance(kwargs[k], na.AbstractVectorArray):
                        kwargs_components[c][k] = kwargs[k].components[c]
                    else:
                        kwargs_components[c][k] = kwargs[k]

            result_expected = tuple(array_normalized.prototype_vector for _ in range(ufunc.nout))
            try:
                for c in components:
                    result_c = ufunc(components[c], components_2[c], **kwargs_components[c])
                    if ufunc.nout == 1:
                        result_c = (result_c,)
                    for i in range(ufunc.nout):
                        result_expected[i].components[c] = result_c[i]
            except Exception as e:
                with pytest.raises(type(e)):
                    ufunc(array, array_2, **kwargs)
                return

            result = ufunc(array, array_2, **kwargs)

            if ufunc.nout == 1:
                out = 0 * np.nan_to_num(result)
            else:
                out = tuple(0 * np.nan_to_num(r) for r in result)

            result_out = ufunc(array, array_2, out=out, **kwargs)

            if ufunc.nout == 1:
                out = (out,)
                result = (result,)
                result_out = (result_out,)

            for i in range(ufunc.nout):
                assert np.all(result[i] == result_expected[i], **kwargs)
                assert np.all(result[i] == result_out[i], **kwargs)
                assert result_out[i] is out[i]


class AbstractTestAbstractExplicitCartesianVectorArray(
    AbstractTestAbstractCartesianVectorArray,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractExplicitVectorArray,
):
    pass


class AbstractTestAbstractImplicitCartesianVectorArray(
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractImplicitVectorArray,
):
    pass


class AbstractTestAbstractCartesianVectorRandomSample(
    AbstractTestAbstractImplicitCartesianVectorArray,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorRandomSample,
):
    pass


class AbstractTestAbstractCartesianVectorUniformRandomSample(
    AbstractTestAbstractCartesianVectorRandomSample,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorUniformRandomSample,
):
    pass


class AbstractTestAbstractCartesianVectorNormalRandomSample(
    AbstractTestAbstractCartesianVectorRandomSample,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorNormalRandomSample,
):
    pass


class AbstractTestAbstractParameterizedCartesianVectorArray(
    AbstractTestAbstractImplicitCartesianVectorArray,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractParameterizedVectorArray,
):
    pass


class AbstractTestAbstractCartesianVectorArrayRange(
    AbstractTestAbstractParameterizedCartesianVectorArray,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorArrayRange,
):
    pass


class AbstractTestAbstractCartesianVectorSpace(
    AbstractTestAbstractParameterizedCartesianVectorArray,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorSpace,
):
    pass


class AbstractTestAbstractCartesianVectorLinearSpace(
    AbstractTestAbstractCartesianVectorSpace,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorLinearSpace,
):
    pass


class AbstractTestAbstractCartesianVectorStratifiedRandomSpace(
    AbstractTestAbstractCartesianVectorLinearSpace,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorStratifiedRandomSpace,
):
    pass


class AbstractTestAbstractCartesianVectorLogarithmicSpace(
    AbstractTestAbstractCartesianVectorSpace,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorLogarithmicSpace,
):
    pass


class AbstractTestAbstractCartesianVectorGeometricSpace(
    AbstractTestAbstractCartesianVectorSpace,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorGeometricSpace,
):
    pass



from typing import Callable
import astropy.units as u
import named_arrays as na
import named_arrays._scalars.scalar_named_array_functions
from . import vectors

__all__ = [
    "RANDOM_FUNCTIONS",
    "HANDLED_FUNCTIONS",
    "random"
]

RANDOM_FUNCTIONS = named_arrays._scalars.scalar_named_array_functions.RANDOM_FUNCTIONS
HANDLED_FUNCTIONS = dict()


def _implements(function: Callable):
    """Register a __named_array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[function] = func
        return func
    return decorator


@_implements(na.arange)
def arange(
        start: float | complex | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
        stop: float | complex | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
        axis: str | na.AbstractVectorArray,
        step: int | na.AbstractVectorArray = 1,
):
    prototype = vectors._prototype(start, stop, axis, step)

    start = vectors._normalize(start, prototype)
    stop = vectors._normalize(stop, prototype)
    axis = vectors._normalize(axis, prototype)
    step = vectors._normalize(step, prototype)

    components_start = start.components
    components_stop = stop.components
    components_axis = axis.components
    components_step = step.components

    components = {
        c: na.arange(
            start=components_start[c],
            stop=components_stop[c],
            axis=components_axis[c],
            step=components_step[c],
        )
        for c in components_start
    }

    return prototype.type_explicit.from_components(components)


def random(
        func: Callable,
        *args: float | u.Quantity |na.AbstractScalar | na.AbstractVectorArray,
        shape_random: None | dict[str, int] = None,
        seed: None | int = None,
        **kwargs: float | u.Quantity |na.AbstractScalar | na.AbstractVectorArray,
):
    try:
        prototype = vectors._prototype(*args, *kwargs.values())
        args = tuple(vectors._normalize(arg, prototype) for arg in args)
        kwargs = {k: vectors._normalize(kwargs[k], prototype) for k in kwargs}
    except na.VectorTypeError:
        return NotImplemented

    components_prototype = prototype.components

    components_args = {c: tuple(arg.components[c] for arg in args) for c in components_prototype}
    components_kwargs = {c: {k: kwargs[k].components[c] for k in kwargs} for c in components_prototype}

    components = {
        c: func(
            *components_args[c],
            shape_random=shape_random,
            seed=seed,
            **components_kwargs[c],
        )
        for c in prototype.components
    }

    return prototype.type_explicit.from_components(components)



from __future__ import annotations
from typing import TypeVar, Generic, ClassVar, Type, Sequence, Callable, Collection, Any
from typing_extensions import Self

import abc
import dataclasses
import numpy as np
import astropy.units as u

import named_arrays as na

__all__ = [
    'UncertainScalarStartT',
    'UncertainScalarStopT',
    'UncertainScalarTypeError',
    'AbstractUncertainScalarArray',
    'UncertainScalarArray',
    'UniformUncertainScalarArray',
    'NormalUncertainScalarArray',
    'UncertainScalarUniformRandomSample',
    'UncertainScalarNormalRandomSample',
    'UncertainScalarPoissionRandomSample',
    'AbstractParameterizedUncertainScalarArray',
    'AbstractUncertainScalarSpace',
    'UncertainScalarLinearSpace',
    'UncertainScalarStratifiedRandomSpace',
    'UncertainScalarLogarithmicSpace',
    'UncertainScalarGeometricSpace',
]

NominalArrayT = TypeVar(
    'NominalArrayT',
    bound=None | float | complex | np.ndarray | u.Quantity | na.AbstractScalarArray,
)
DistributionArrayT = TypeVar(
    'DistributionArrayT',
    bound=None | float | complex | np.ndarray | u.Quantity | na.AbstractScalarArray,
)
WidthT = TypeVar('WidthT', bound=int | float | np.ndarray | u.Quantity | na.AbstractScalarArray)
UncertainScalarStartT = TypeVar("UncertainScalarStartT", bound=float | u.Quantity | na.AbstractScalar)
UncertainScalarStopT = TypeVar("UncertainScalarStopT", bound=float | u.Quantity | na.AbstractScalar)
UncertainScalarCenterT = TypeVar("UncertainScalarCenterT", bound=float | u.Quantity | na.AbstractScalar)
UncertainScalarWidthT = TypeVar("UncertainScalarWidthT", bound=float | u.Quantity | na.AbstractScalar)

_axis_distribution_default = "_distribution"
_num_distribution_default = 11


class UncertainScalarTypeError(TypeError):
    pass


def _normalize(a: float | u.Quantity | na.AbstractScalar):
    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractScalar):
            if isinstance(a, na.AbstractUncertainScalarArray):
                result = a
            else:
                result = na.UncertainScalarArray(a, a)
        else:
            return UncertainScalarTypeError
    else:
        result = na.UncertainScalarArray(a, a)

    return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractUncertainScalarArray(
    na.AbstractScalar
):
    __named_array_priority__: ClassVar[int] = 10 * na.AbstractScalarArray.__named_array_priority__

    axis_distribution: ClassVar[str] = "_distribution"

    @property
    def type_explicit(self) -> Type[UncertainScalarArray]:
        return UncertainScalarArray

    @property
    def type_abstract(self) -> Type[AbstractUncertainScalarArray]:
        return AbstractUncertainScalarArray

    @property
    @abc.abstractmethod
    def nominal(self) -> float | complex | u.Quantity | na.AbstractScalarArray:
        """
        Nominal value of the array.
        """

    @property
    @abc.abstractmethod
    def distribution(self) -> na.AbstractScalarArray:
        """
        Distribution of possible values of the array.
        """

    @property
    @abc.abstractmethod
    def num_distribution(self) -> int:
        """
        Number samples along :attr:`axis_distribution`.
        """

    @property
    def shape_distribution(self) -> dict[str, int]:
        return na.shape_broadcasted(self.nominal, self.distribution)

    @property
    def dtype(self) -> np.dtype:
        return np.promote_types(
            na.get_dtype(self.nominal),
            na.get_dtype(self.distribution),
        )

    @property
    def value(self) -> UncertainScalarArray:
        return self.type_explicit(
            nominal=na.value(self.nominal),
            distribution=na.value(self.distribution),
        )

    @property
    def unit(self: Self) -> None | u.Unit:
        return na.unit(self.nominal)

    def astype(
            self,
            dtype: str | np.dtype | Type,
            order: str = 'K',
            casting='unsafe',
            subok: bool = True,
            copy: bool = True,
    ) -> UncertainScalarArray:
        return UncertainScalarArray(
            nominal=na.as_named_array(self.nominal).astype(
                dtype=dtype,
                order=order,
                casting=casting,
                subok=subok,
                copy=copy,
            ),
            distribution=self.distribution.astype(
                dtype=dtype,
                order=order,
                casting=casting,
                subok=subok,
                copy=copy,
            ),
        )

    def to(self, unit: u.UnitBase) -> UncertainScalarArray:
        return UncertainScalarArray(
            nominal=na.as_named_array(self.nominal).to(unit),
            distribution=self.distribution.to(unit),
        )

    def add_axes(self, axes: str | Sequence[str]) -> UncertainScalarArray:
        return UncertainScalarArray(
            nominal=na.as_named_array(self.nominal).add_axes(axes),
            distribution=self.distribution.add_axes(axes),
        )

    def combine_axes(
            self,
            axes: None | Sequence[str] = None,
            axis_new: None | str =None,
    ) -> UncertainScalarArray:

        shape = self.shape

        if axes is None:
            axes = tuple(self.shape)

        shape_base = {ax: shape[ax] for ax in shape if ax in axes}

        nominal = na.broadcast_to(self.nominal, na.shape(self.nominal) | shape_base)
        distribution = na.broadcast_to(self.distribution, na.shape(self.distribution) | shape_base)

        return UncertainScalarArray(
            nominal=nominal.combine_axes(axes=axes, axis_new=axis_new),
            distribution=distribution.combine_axes(axes=axes, axis_new=axis_new),
        )

    def _getitem(
            self,
            item: dict[str, int | slice | na.AbstractScalar] | na.AbstractScalar,
    ):
        array = self.explicit
        shape_array = array.shape
        shape_array_distribution = array.shape_distribution

        nominal = na.as_named_array(array.nominal)
        distribution = na.as_named_array(array.distribution)

        if isinstance(item, na.AbstractArray):
            item = item.explicit
            if isinstance(item, AbstractUncertainScalarArray):
                item_nominal = item_distribution = item.nominal & np.all(item.distribution, axis=self.axis_distribution)
            elif isinstance(item, na.AbstractScalarArray):
                item_nominal = item_distribution = item
            else:
                return NotImplemented

            shape_item = na.broadcast_shapes(item_nominal.shape, item_distribution.shape)

            if not set(shape_item).issubset(shape_array_distribution):
                raise ValueError(
                    f"the axes in item, {tuple(shape_item)}, must be a subset of the axes in array, "
                    f"{tuple(shape_array_distribution)}"
                )

            if not all(shape_item[ax] == shape_array_distribution[ax] for ax in shape_item):
                raise ValueError(
                    f"the shape of item, {shape_item}, must be consistent with the shape of the array, "
                    f"{shape_array_distribution}"
                )

            shape_nominal = na.broadcast_shapes(nominal.shape, item_nominal.shape)
            shape_distribution = na.broadcast_shapes(distribution.shape, item_distribution.shape)

            nominal = na.broadcast_to(nominal, shape_nominal)
            distribution = na.broadcast_to(distribution, shape_distribution)

        elif isinstance(item, dict):

            if not set(item).issubset(shape_array_distribution):
                raise ValueError(f"the axes in item, {tuple(item)}, must be a subset of the axes in the array, {array.axes}")

            item_nominal = dict()
            item_distribution = dict()
            for ax in item:
                if isinstance(item[ax], na.AbstractArray):
                    if isinstance(item[ax], AbstractUncertainScalarArray):
                        item_nominal[ax] = item[ax].nominal
                        item_distribution[ax] = item[ax].distribution
                    elif isinstance(item[ax], na.AbstractScalarArray):
                        item_nominal[ax] = item_distribution[ax] = item[ax]
                    else:
                        return NotImplemented
                elif isinstance(item[ax], (int, slice)):
                    item_nominal[ax] = item_distribution[ax] = item[ax]
                else:
                    return NotImplemented

                if ax not in nominal.axes:
                    item_nominal.pop(ax)
                if ax not in distribution.axes:
                    item_distribution.pop(ax)

        else:
            return NotImplemented

        result = UncertainScalarArray(
            nominal=nominal[item_nominal],
            distribution=distribution[item_distribution],
        )

        return result

    def _getitem_reversed(
            self,
            array: na.ScalarArray,
            item: dict[str, int | slice | AbstractUncertainScalarArray] | AbstractUncertainScalarArray
    ):
        if isinstance(array, AbstractUncertainScalarArray):
            pass
        elif isinstance(array, na.AbstractScalarArray):
            if isinstance(item, dict):
                num_distribution = item[self.axis_distribution].distribution.max().ndarray + 1
            else:
                num_distribution = item.num_distribution
            shape_distribution = na.broadcast_shapes(array.shape, {self.axis_distribution: num_distribution})
            array = UncertainScalarArray(
                nominal=array,
                distribution=array.broadcast_to(shape_distribution),
            )
        else:
            return NotImplemented

        return array._getitem(item)

    def __bool__(self):
        result = super().__bool__()
        nominal = bool(self.nominal)
        distribution = self.distribution
        if self.axis_distribution in distribution.shape:
            distribution = np.all(distribution, axis=self.axis_distribution)
        distribution = bool(distribution)
        return result and nominal and distribution

    def __mul__(self, other: na.ArrayLike | u.Unit) -> UncertainScalarArray:
        if isinstance(other, u.UnitBase):
            return UncertainScalarArray(
                nominal=self.nominal * other,
                distribution=self.distribution * other,
            )
        else:
            return super().__mul__(other)

    def __lshift__(self, other: na.ArrayLike | u.Unit) -> UncertainScalarArray:
        if isinstance(other, u.UnitBase):
            return UncertainScalarArray(
                nominal=self.nominal << other,
                distribution=self.distribution << other,
            )
        else:
            return super().__lshift__(other)

    def __truediv__(self, other: na.ArrayLike | u.Unit) -> UncertainScalarArray:
        if isinstance(other, u.UnitBase):
            return UncertainScalarArray(
                nominal=self.nominal / other,
                distribution=self.distribution / other,
            )
        else:
            return super().__truediv__(other)

    def __array_ufunc__(
            self,
            function: np.ufunc,
            method: str,
            *inputs,
            **kwargs,
    ) -> None | UncertainScalarArray | tuple[UncertainScalarArray, ...]:

        result = super().__array_ufunc__(function, method, *inputs, **kwargs)
        if result is not NotImplemented:
            return result

        nout = function.nout

        inputs_nominal = []
        inputs_distribution = []
        for inp in inputs:
            if isinstance(inp, na.AbstractArray):
                if isinstance(inp, AbstractUncertainScalarArray):
                    inp_nominal = inp.nominal
                    inp_distribution = inp.distribution
                elif isinstance(inp, na.AbstractScalarArray):
                    inp_nominal = inp_distribution = inp
                else:
                    return NotImplemented
            else:
                inp_nominal = inp_distribution = inp
            inputs_nominal.append(inp_nominal)
            inputs_distribution.append(inp_distribution)

        kwargs_nominal = dict()
        kwargs_distribution = dict()

        if "where" in kwargs:
            where = kwargs.pop("where")
            if isinstance(where, na.AbstractArray):
                if isinstance(where, na.AbstractScalar):
                    if isinstance(where, AbstractUncertainScalarArray):
                        where_nominal = where.nominal
                        where_distribution = where.distribution
                    else:
                        where_nominal = where_distribution = where
                else:
                    return NotImplemented
            else:
                where_nominal = where_distribution = where
            kwargs_nominal["where"] = where_nominal
            kwargs_distribution["where"] = where_distribution

        if "out" in kwargs:
            out = kwargs.pop("out")
            out_nominal = list()
            out_distribution = list()
            for o in out:
                if o is not None:
                    if isinstance(o, UncertainScalarArray):
                        types = (np.ndarray, na.AbstractArray)
                        o_nominal = o.nominal if isinstance(o.nominal, types) else None
                        o_distribution = o.distribution if isinstance(o.distribution, types) else None
                    else:
                        raise ValueError(
                            f"`out` must be `None` or an instance of `{self.type_explicit}`, "
                            f"got {tuple(type(x) for x in out)}"
                        )
                else:
                    o_nominal = o_distribution = None
                out_nominal.append(o_nominal)
                out_distribution.append(o_distribution)
            if nout == 1:
                out_nominal = out_nominal[0]
                out_distribution = out_distribution[0]
            else:
                out_nominal = tuple(out_nominal)
                out_distribution = tuple(out_distribution)
            kwargs_nominal["out"] = out_nominal
            kwargs_distribution["out"] = out_distribution
        else:
            out = (None, ) * nout

        result_nominal = getattr(function, method)(*inputs_nominal, **kwargs_nominal, **kwargs)
        result_distribution = getattr(function, method)(*inputs_distribution, **kwargs_distribution, **kwargs)

        if nout == 1:
            result_nominal = (result_nominal, )
            result_distribution = (result_distribution, )

        result = list(
            UncertainScalarArray(result_nominal[i], result_distribution[i])
            for i in range(nout)
        )

        for i in range(nout):
            if out[i] is not None:
                out[i].nominal = result[i].nominal
                out[i].distribution = result[i].distribution
                result[i] = out[i]

        if nout == 1:
            result = result[0]
        else:
            result = tuple(result)
        return result

    def __array_function__(
            self: Self,
            func: Callable,
            types: Collection,
            args: tuple,
            kwargs: dict[str, Any],
    ):
        result = super().__array_function__(func=func, types=types, args=args, kwargs=kwargs)
        if result is not NotImplemented:
            return result

        from . import uncertainties_array_functions

        if func in uncertainties_array_functions.ARRAY_CREATION_LIKE_FUNCTIONS:
            return uncertainties_array_functions.array_function_array_creation_like(func, *args, **kwargs)

        if func in uncertainties_array_functions.SEQUENCE_FUNCTIONS:
            return uncertainties_array_functions.array_function_sequence(func, *args, **kwargs)

        if func in uncertainties_array_functions.DEFAULT_FUNCTIONS:
            return uncertainties_array_functions.array_function_default(func, *args, **kwargs)

        if func in uncertainties_array_functions.PERCENTILE_LIKE_FUNCTIONS:
            return uncertainties_array_functions.array_function_percentile_like(func, *args, **kwargs)

        if func in uncertainties_array_functions.ARG_REDUCE_FUNCTIONS:
            return uncertainties_array_functions.array_function_arg_reduce(func, *args, **kwargs)

        if func in uncertainties_array_functions.FFT_LIKE_FUNCTIONS:
            return uncertainties_array_functions.array_function_fft_like(func, *args, **kwargs)

        if func in uncertainties_array_functions.FFTN_LIKE_FUNCTIONS:
            return uncertainties_array_functions.array_function_fftn_like(func, *args, **kwargs)

        if func in uncertainties_array_functions.STACK_LIKE_FUNCTIONS:
            return uncertainties_array_functions.array_function_stack_like(func, *args, **kwargs)

        if func in uncertainties_array_functions.HANDLED_FUNCTIONS:
            return uncertainties_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    def __named_array_function__(self, func, *args, **kwargs):
        result = super().__named_array_function__(func, *args, **kwargs)
        if result is not NotImplemented:
            return result

        from . import uncertainties_named_array_functions

        if func in uncertainties_named_array_functions.RANDOM_FUNCTIONS:
            return uncertainties_named_array_functions.random(func=func, *args, **kwargs)

        if func in uncertainties_named_array_functions.PLT_PLOT_LIKE_FUNCTIONS:
            return uncertainties_named_array_functions.plt_plot_like(func, *args, **kwargs)

        if func in uncertainties_named_array_functions.HANDLED_FUNCTIONS:
            return uncertainties_named_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    @property
    def broadcasted(self) -> na.UncertainScalarArray:
        a = self.explicit
        return na.UncertainScalarArray(
            nominal=na.broadcast_to(a.nominal, a.shape),
            distribution=na.broadcast_to(a.distribution, a.shape_distribution),
        )


@dataclasses.dataclass(eq=False, repr=False)
class UncertainScalarArray(
    AbstractUncertainScalarArray,
    na.AbstractExplicitArray,
    Generic[NominalArrayT, DistributionArrayT],
):
    nominal: NominalArrayT = dataclasses.MISSING
    distribution: DistributionArrayT = dataclasses.MISSING

    def __post_init__(self):
        if self.axis_distribution in na.shape(self.nominal):
            raise ValueError(
                f"`axis_distribution`, '{self.axis_distribution}' should not be in `nominal` array with "
                f"shape {na.shape(self.nominal)}"
            )

    @property
    def num_distribution(self: Self) -> int:
        return self.distribution.shape[self.axis_distribution]

    @property
    def axes(self: Self) -> tuple[str, ...]:
        return tuple(self.shape.keys())

    @property
    def shape(self) -> dict[str, int]:
        shape = self.shape_distribution
        if self.axis_distribution in shape:
            shape.pop(self.axis_distribution)
        return shape

    @property
    def ndim(self: Self) -> int:
        return len(self.shape)

    @property
    def size(self: Self) -> int:
        return int(np.array(tuple(self.shape.values())).prod())

    @property
    def explicit(self) -> Self:
        return self.copy_shallow()

    @property
    def centers(self) -> Self:
        return self

    def __setitem__(
            self,
            item: dict[str, int | slice | na.AbstractScalar] | na.AbstractScalar,
            value: int | float | u.Quantity | na.AbstractScalar,
    ):
        shape_self = self.shape

        if isinstance(item, na.AbstractArray):

            item = item.explicit
            if not set(item.shape).issubset(shape_self):
                raise ValueError(
                    f"if `item` is an instance of `{na.AbstractArray.__name__}`, "
                    f"`item.axes`, {item.axes}, should be a subset of `self.axes`, {self.axes}"
                )

            if isinstance(item, na.AbstractUncertainScalarArray):
                item_nominal = item.nominal
                item_distribution = item.distribution
            elif isinstance(item, na.AbstractScalarArray):
                item_nominal = item_distribution = item
            else:
                raise TypeError(
                    f"if `item` is an instance of `{na.AbstractArray.__name__}`, "
                    f"it must be an instance of `{na.AbstractScalar.__name__}`, "
                    f"got `{type(item)}`"
                )
        elif isinstance(item, dict):

            if not set(item).issubset(shape_self):
                raise ValueError(
                    f"if `item` is a `{dict.__name__}`, the keys in `item`, {tuple(item)}, "
                    f"must be a subset of `self.axes`, {self.axes}"
                )

            item_nominal = dict()
            item_distribution = dict()
            for axis in item:
                item_axis = item[axis]
                if isinstance(item_axis, na.AbstractArray):
                    if isinstance(item_axis, na.AbstractUncertainScalarArray):
                        item_nominal[axis] = item_axis.nominal
                        item_distribution[axis] = item_axis.distribution
                    elif isinstance(item_axis, na.AbstractScalarArray):
                        item_nominal[axis] = item_distribution[axis] = item_axis
                    else:
                        raise TypeError(
                            f"if a value in `item` is an instance of `{na.AbstractArray.__name__}`, "
                            f"it must be an instance of `{na.AbstractScalar.__name__}`, "
                            f"got `{type(item_axis)}`"
                        )
                else:
                    item_nominal[axis] = item_distribution[axis] = item_axis

        else:
            raise TypeError(
                f"`item` must be an instance of `{na.AbstractArray.__name__}` or {dict.__name__}, "
                f"got `{type(item)}`"
            )

        if isinstance(value, na.AbstractArray):
            if isinstance(value, na.AbstractUncertainScalarArray):
                value_nominal = value.nominal
                value_distribution = value.distribution
            elif isinstance(value, na.AbstractScalarArray):
                value_nominal = value_distribution = value
            else:
                raise TypeError(
                    f"if `value` is an instance of `{na.AbstractArray.__name__}`, "
                    f"it must be an instance of `{na.AbstractScalar.__name__}`, "
                    f"got {type(value)}"
                )
        else:
            value_nominal = value_distribution = value

        self.nominal[item_nominal] = value_nominal
        self.distribution[item_distribution] = value_distribution


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitUncertainScalarArray(
    AbstractUncertainScalarArray,
    na.AbstractImplicitArray,
):

    def _attr_normalized(self, name: str) -> UncertainScalarArray:

        attr = getattr(self, name)

        if isinstance(attr, na.AbstractArray):
            if isinstance(attr, na.AbstractScalar):
                if isinstance(attr, na.AbstractUncertainScalarArray):
                    result = attr
                else:
                    result = UncertainScalarArray(attr, attr)
            else:
                raise TypeError(
                    f"if `{name}` is an instance of `AbstractArray`, it must be an instance of `AbstractScalar`, "
                    f"got {type(attr)}"
                )
        else:
            result = UncertainScalarArray(attr, attr)

        return result


@dataclasses.dataclass(eq=False, repr=False)
class UniformUncertainScalarArray(
    AbstractImplicitUncertainScalarArray,
    na.AbstractRandomMixin,
    Generic[NominalArrayT, WidthT],
):
    nominal: NominalArrayT = dataclasses.MISSING
    width: WidthT = dataclasses.MISSING
    num_distribution: int = _num_distribution_default
    seed: None | int = None

    @property
    def distribution(self: Self) -> na.ScalarUniformRandomSample:
        return na.ScalarUniformRandomSample(
            start=self.nominal - self.width,
            stop=self.nominal + self.width,
            shape_random={self.axis_distribution: self.num_distribution},
            seed=self.seed,
        )

    @property
    def explicit(self) -> UncertainScalarArray:
        return UncertainScalarArray(
            nominal=na.explicit(self.nominal),
            distribution=na.explicit(self.distribution),
        )

    @property
    def centers(self) -> Self:
        return self


@dataclasses.dataclass(eq=False, repr=False)
class NormalUncertainScalarArray(
    AbstractImplicitUncertainScalarArray,
    na.AbstractRandomMixin,
    Generic[NominalArrayT, WidthT],
):
    nominal: NominalArrayT = dataclasses.MISSING
    width: WidthT = dataclasses.MISSING
    num_distribution: int = _num_distribution_default
    seed: None | int = None

    @property
    def distribution(self: Self) -> na.ScalarNormalRandomSample:
        return na.ScalarNormalRandomSample(
            center=self.nominal,
            width=self.width,
            shape_random={self.axis_distribution: self.num_distribution},
            seed=self.seed,
        )

    @property
    def explicit(self) -> UncertainScalarArray:
        return UncertainScalarArray(
            nominal=na.explicit(self.nominal),
            distribution=na.explicit(self.distribution),
        )

    @property
    def centers(self) -> Self:
        return self


@dataclasses.dataclass(eq=False, repr=False)
class AbstractUncertainScalarRandomSample(
    AbstractImplicitUncertainScalarArray,
    na.AbstractRandomSample,
):
    @property
    def nominal(self) -> float | u.Quantity | na.AbstractScalarArray:
        return self.explicit.nominal

    @property
    def distribution(self) -> na.AbstractScalarArray:
        return self.explicit.distribution

    @property
    def num_distribution(self) -> int:
        return self.explicit.num_distribution


@dataclasses.dataclass(eq=False, repr=False)
class UncertainScalarUniformRandomSample(
    AbstractUncertainScalarRandomSample,
    na.AbstractUniformRandomSample[UncertainScalarStartT, UncertainScalarStopT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class UncertainScalarNormalRandomSample(
    AbstractUncertainScalarRandomSample,
    na.AbstractNormalRandomSample[UncertainScalarCenterT, UncertainScalarWidthT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class UncertainScalarPoissionRandomSample(
    AbstractUncertainScalarRandomSample,
    na.AbstractPoissonRandomSample[UncertainScalarCenterT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedUncertainScalarArray(
    AbstractImplicitUncertainScalarArray,
    na.AbstractParameterizedArray,
):
    @property
    def nominal(self) -> float | u.Quantity | na.AbstractScalarArray:
        return self.explicit.nominal

    @property
    def distribution(self) -> na.AbstractScalarArray:
        return self.explicit.distribution

    @property
    def num_distribution(self) -> int:
        return self.explicit.num_distribution


@dataclasses.dataclass(eq=False, repr=False)
class AbstractUncertainScalarSpace(
    AbstractParameterizedUncertainScalarArray,
    na.AbstractSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class UncertainScalarLinearSpace(
    AbstractUncertainScalarSpace,
    na.AbstractLinearSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class UncertainScalarStratifiedRandomSpace(
    UncertainScalarLinearSpace,
    na.AbstractStratifiedRandomSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class UncertainScalarLogarithmicSpace(
    AbstractUncertainScalarSpace,
    na.AbstractLogarithmicSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class UncertainScalarGeometricSpace(
    AbstractUncertainScalarSpace,
    na.AbstractGeometricSpace,
):
    pass



from __future__ import annotations
from typing import TypeVar, Generic, Type
import abc
import dataclasses
import numpy as np
import named_arrays as na

__all__ = [
    'AbstractCartesian3dMatrixArray',
    'Cartesian3dMatrixArray',
    'AbstractImplicitCartesian3dMatrixArray',
    'Cartesian3dIdentityMatrixArray',
    'AbstractCartesian3dRotationMatrixArray',
    'Cartesian3dXRotationMatrixArray',
    'Cartesian3dYRotationMatrixArray',
    'Cartesian3dZRotationMatrixArray',
]

XT = TypeVar('XT', bound=na.AbstractVectorArray)
YT = TypeVar('YT', bound=na.AbstractVectorArray)
ZT = TypeVar('ZT', bound=na.AbstractVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian3dMatrixArray(
    na.AbstractCartesian3dVectorArray,
    na.AbstractCartesian2dMatrixArray,
):
    @property
    @abc.abstractmethod
    def z(self) -> na.AbstractVectorArray:
        """
        The `z` row of the matrix.
        """

    @property
    def type_abstract(self) -> Type[AbstractCartesian3dMatrixArray]:
        return AbstractCartesian3dMatrixArray

    @property
    def type_explicit(self) -> Type[Cartesian3dMatrixArray]:
        return Cartesian3dMatrixArray

    @property
    def cartesian_nd(self):
        return NotImplementedError

    @property
    def type_vector(self) -> Type[na.Cartesian3dVectorArray]:
        return na.Cartesian3dVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        if not self.is_square:
            raise ValueError("can only compute determinant of square matrices")
        a, b, c = self.x.components.values()
        d, e, f = self.y.components.values()
        g, h, i = self.z.components.values()
        return (a * e * i) + (b * f * g) + (c * d * h) - (c * e * g) - (b * d * i) - (a * f * h)

    @property
    def inverse(self) -> na.AbstractMatrixArray:
        if not self.is_square:
            raise ValueError("can only compute inverse of square matrices")
        type_matrix = self.x.type_matrix
        type_row = self.type_vector
        c1, c2, c3 = self.x.components
        a, b, c = self.x.components.values()
        d, e, f = self.y.components.values()
        g, h, i = self.z.components.values()
        result = type_matrix.from_components({
            c1: type_row(
                x=(e * i - f * h),
                y=-(b * i - c * h),
                z=(b * f - c * e),
            ),
            c2: type_row(
                x=-(d * i - f * g),
                y=(a * i - c * g),
                z=-(a * f - c * d),
            ),
            c3: type_row(
                x=(d * h - e * g),
                y=-(a * h - b * g),
                z=(a * e - b * d),
            )
        })
        result = result / self.determinant
        return result


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dMatrixArray(
    na.Cartesian3dVectorArray,
    AbstractCartesian3dMatrixArray,
    na.Cartesian2dMatrixArray,
    Generic[XT, YT, ZT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitCartesian3dMatrixArray(
    AbstractCartesian3dMatrixArray,
    na.AbstractImplicitCartesian2dMatrixArray,
    na.AbstractImplicitCartesian3dVectorArray,
):

    @property
    def z(self) -> na.ArrayLike:
        return self.explicit.z


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dIdentityMatrixArray(
    AbstractImplicitCartesian3dMatrixArray,
    na.Cartesian2dIdentityMatrixArray,
):

    @property
    def explicit(self) -> na.Cartesian3dMatrixArray:
        return Cartesian3dMatrixArray(
            x=na.Cartesian3dVectorArray(x=1, y=0, z=0),
            y=na.Cartesian3dVectorArray(x=0, y=1, z=0),
            z=na.Cartesian3dVectorArray(x=0, y=0, z=1),
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian3dRotationMatrixArray(
    AbstractImplicitCartesian3dMatrixArray,
    na.AbstractCartesian2dRotationMatrixArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dXRotationMatrixArray(
    AbstractCartesian3dRotationMatrixArray,
):
    angle: na.ScalarLike = dataclasses.MISSING

    @property
    def explicit(self) -> Cartesian3dMatrixArray:
        a = self.angle
        return Cartesian3dMatrixArray(
            x=na.Cartesian3dVectorArray(x=1, y=0, z=0),
            y=na.Cartesian3dVectorArray(x=0, y=np.cos(a), z=-np.sin(a)),
            z=na.Cartesian3dVectorArray(x=0, y=np.sin(a), z=np.cos(a)),
        )


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dYRotationMatrixArray(
    AbstractCartesian3dRotationMatrixArray,
):
    angle: na.ScalarLike = dataclasses.MISSING

    @property
    def explicit(self) -> Cartesian3dMatrixArray:
        a = self.angle
        return Cartesian3dMatrixArray(
            x=na.Cartesian3dVectorArray(x=np.cos(a), y=0, z=np.sin(a)),
            y=na.Cartesian3dVectorArray(x=0, y=1, z=0),
            z=na.Cartesian3dVectorArray(x=-np.sin(a), y=0, z=np.cos(a)),
        )


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dZRotationMatrixArray(
    AbstractCartesian3dRotationMatrixArray,
):
    angle: na.ScalarLike = dataclasses.MISSING

    @property
    def explicit(self) -> Cartesian3dMatrixArray:
        a = self.angle
        return Cartesian3dMatrixArray(
            x=na.Cartesian3dVectorArray(x=np.cos(a), y=-np.sin(a), z=0),
            y=na.Cartesian3dVectorArray(x=np.sin(a), y=np.cos(a), z=0),
            z=na.Cartesian3dVectorArray(x=0, y=0, z=1),
        )



from __future__ import annotations
from typing import TypeVar, Type, Generic
import numpy as np
from typing_extensions import Self
import abc
import dataclasses
import astropy.units as u
import named_arrays as na

__all__ = [
    'AbstractCartesian2dVectorArray',
    'Cartesian2dVectorArray',
    'AbstractImplicitCartesian2dVectorArray',
    'AbstractCartesian2dVectorRandomSample',
    'Cartesian2dVectorUniformRandomSample',
    'Cartesian2dVectorNormalRandomSample',
    'AbstractParameterizedCartesian2dVectorArray',
    'Cartesian2dVectorArrayRange',
    'AbstractCartesian2dVectorSpace',
    'Cartesian2dVectorLinearSpace',
    'Cartesian2dVectorStratifiedRandomSpace',
    'Cartesian2dVectorLogarithmicSpace',
    'Cartesian2dVectorGeometricSpace',
]

XT = TypeVar('XT', bound=na.ArrayLike)
YT = TypeVar('YT', bound=na.ArrayLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian2dVectorArray(
    na.AbstractCartesianVectorArray,
):

    @property
    @abc.abstractmethod
    def x(self: Self) -> na.ArrayLike:
        """
        The `x` component of the vector.
        """

    @property
    @abc.abstractmethod
    def y(self: Self) -> na.ArrayLike:
        """
        The `y` component of the vector.
        """

    @property
    def type_abstract(self: Self) -> Type[AbstractCartesian2dVectorArray]:
        return AbstractCartesian2dVectorArray

    @property
    def type_explicit(self: Self) -> Type[Cartesian2dVectorArray]:
        return Cartesian2dVectorArray

    @property
    def type_matrix(self) -> Type[na.Cartesian2dMatrixArray]:
        return na.Cartesian2dMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dVectorArray(
    AbstractCartesian2dVectorArray,
    na.AbstractExplicitCartesianVectorArray,
    Generic[XT, YT],
):
    x: XT = 0
    y: YT = 0

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> Cartesian2dVectorArray:
        result = super().from_scalar(scalar, like=like)
        if result is not NotImplemented:
            return result

        return cls(x=scalar, y=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitCartesian2dVectorArray(
    AbstractCartesian2dVectorArray,
    na.AbstractImplicitCartesianVectorArray,
):

    @property
    def x(self) -> na.ArrayLike:
        return self.explicit.x

    @property
    def y(self) -> na.ArrayLike:
        return self.explicit.y


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian2dVectorRandomSample(
    AbstractImplicitCartesian2dVectorArray,
    na.AbstractCartesianVectorRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dVectorUniformRandomSample(
    AbstractCartesian2dVectorRandomSample,
    na.AbstractCartesianVectorUniformRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dVectorNormalRandomSample(
    AbstractCartesian2dVectorRandomSample,
    na.AbstractCartesianVectorNormalRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedCartesian2dVectorArray(
    AbstractImplicitCartesian2dVectorArray,
    na.AbstractParameterizedCartesianVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dVectorArrayRange(
    AbstractParameterizedCartesian2dVectorArray,
    na.AbstractCartesianVectorArrayRange,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian2dVectorSpace(
    AbstractParameterizedCartesian2dVectorArray,
    na.AbstractCartesianVectorSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dVectorLinearSpace(
    AbstractCartesian2dVectorSpace,
    na.AbstractCartesianVectorLinearSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dVectorStratifiedRandomSpace(
    Cartesian2dVectorLinearSpace,
    na.AbstractCartesianVectorStratifiedRandomSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dVectorLogarithmicSpace(
    AbstractCartesian2dVectorSpace,
    na.AbstractCartesianVectorLogarithmicSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dVectorGeometricSpace(
    AbstractCartesian2dVectorSpace,
    na.AbstractCartesianVectorGeometricSpace,
):
    pass



from __future__ import annotations
from typing import TypeVar, Generic, Type, ClassVar, Sequence, Callable, Collection, Any
import abc
import dataclasses
import numpy as np
import astropy.units as u
import astropy.visualization
import named_arrays as na

__all__ = [
    "InputValueError",
    "AbstractFunctionArray",
    "FunctionArray",
]

InputsT = TypeVar("InputsT", bound=na.AbstractArray)
OutputsT = TypeVar("OutputsT", bound=na.AbstractArray)


class InputValueError(ValueError):
    """
    Exception raised when the inputs of two functions do not match
    """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractFunctionArray(
    na.AbstractArray,
):

    @property
    @abc.abstractmethod
    def inputs(self) -> na.AbstractArray:
        """
        The arrays representing the inputs to this function.
        """

    @property
    @abc.abstractmethod
    def outputs(self) -> na.AbstractArray:
        """
        The arrays representing the outputs of this function.
        """

    __named_array_priority__: ClassVar[float] = 100 * na.AbstractVectorArray.__named_array_priority__

    @property
    def type_abstract(self) -> Type[AbstractFunctionArray]:
        return AbstractFunctionArray

    @property
    def type_explicit(self) -> Type[FunctionArray]:
        return FunctionArray

    @property
    def value(self) -> FunctionArray:
        return self.type_explicit(
            inputs=self.inputs,
            outputs=self.outputs.value,
        )

    @property
    def centers(self) -> FunctionArray:
        return self.type_explicit(
            inputs=self.inputs.centers,
            outputs=self.outputs.centers,
        )

    def astype(
            self,
            dtype: str | np.dtype | type,
            order: str = 'K',
            casting='unsafe',
            subok: bool = True,
            copy: bool = True,
    ) -> FunctionArray:

        return self.type_explicit(
            inputs=self.inputs,
            outputs=self.outputs.astype(
                dtype=dtype,
                order=order,
                casting=casting,
                subok=subok,
                copy=copy,
            )
        )

    def to(self, unit: u.UnitBase) -> FunctionArray:
        return self.type_explicit(
            inputs=self.inputs,
            outputs=self.outputs.to(unit),
        )

    @property
    def length(self) -> FunctionArray:
        return self.type_explicit(
            inputs=self.inputs,
            outputs=self.outputs.length,
        )

    def add_axes(self, axes: str | Sequence[str]) -> FunctionArray:
        return self.type_explicit(
            inputs=self.inputs,
            outputs=self.outputs.add_axes(axes),
        )

    def combine_axes(
            self,
            axes: Sequence[str] = None,
            axis_new: str = None,
    ) -> FunctionArray:

        a = self.explicit
        shape = a.shape

        axes = tuple(shape) if axes is None else axes

        shape_base = {ax: shape[ax] for ax in shape if ax in axes}

        inputs = a.inputs.broadcast_to(na.broadcast_shapes(a.inputs.shape, shape_base))
        outputs = a.outputs.broadcast_to(na.broadcast_shapes(a.outputs.shape, shape_base))
        return self.type_explicit(
            inputs=inputs.combine_axes(axes=axes, axis_new=axis_new),
            outputs=outputs.combine_axes(axes=axes, axis_new=axis_new),
        )

    def _getitem(
            self,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractFunctionArray,
    ) -> FunctionArray:

        array = self.explicit
        inputs = array.inputs
        outputs = array.outputs

        shape = array.shape
        shape_inputs = inputs.shape
        shape_outputs = outputs.shape

        if isinstance(item, na.AbstractFunctionArray):
            if np.any(item.inputs != self.inputs):
                raise ValueError("boolean advanced index does not have the same inputs as the array")

            item_inputs = item.outputs
            item_outputs = item.outputs

            shape_item_inputs = item_inputs.shape
            shape_item_outputs = item_outputs.shape

        elif isinstance(item, dict):

            item_inputs = dict()
            item_outputs = dict()
            for ax in item:
                item_ax = item[ax]
                if isinstance(item_ax, na.AbstractFunctionArray):
                    item_inputs[ax] = item_ax.inputs
                    item_outputs[ax] = item_ax.outputs
                else:
                    item_inputs[ax] = item_outputs[ax] = item_ax

            shape_item_inputs = {ax: shape[ax] for ax in item_inputs if ax in shape}
            shape_item_outputs = {ax: shape[ax] for ax in item_outputs if ax in shape}

        else:
            return NotImplemented

        inputs = na.broadcast_to(inputs, na.broadcast_shapes(shape_inputs, shape_item_inputs))
        outputs = na.broadcast_to(outputs, na.broadcast_shapes(shape_outputs, shape_item_outputs))

        return self.type_explicit(
            inputs=inputs[item_inputs],
            outputs=outputs[item_outputs],
        )

    def _getitem_reversed(
            self,
            array: na.AbstractArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractFunctionArray
    ):
        if isinstance(array, (na.AbstractScalar, na.AbstractVectorArray)):
            array = na.FunctionArray(array, array)
        else:
            return NotImplemented
        return array._getitem(item)

    def __bool__(self) -> bool:
        result = super().__bool__()
        return result and bool(self.outputs)

    def __mul__(self, other: na.ArrayLike | u.UnitBase) -> FunctionArray:
        if isinstance(other, u.UnitBase):
            return self.type_explicit(
                inputs=self.inputs,
                outputs=self.outputs * other,
            )
        else:
            return super().__mul__(other)

    def __lshift__(self, other: na.ArrayLike | u.UnitBase) -> FunctionArray:
        if isinstance(other, u.UnitBase):
            return self.type_explicit(
                inputs=self.inputs,
                outputs=self.outputs << other,
            )
        else:
            return super().__lshift__(other)

    def __truediv__(self, other: na.ArrayLike | u.UnitBase) -> FunctionArray:
        if isinstance(other, u.UnitBase):
            return self.type_explicit(
                inputs=self.inputs,
                outputs=self.outputs / other,
            )
        else:
            return super().__truediv__(other)

    def __array_matmul__(
            self,
            x1: float | u.Quantity | FunctionArray,
            x2: float | u.Quantity | FunctionArray,
            out: None | na.FunctionArray = None,
            **kwargs,
    ) -> FunctionArray:

        if isinstance(x1, na.AbstractArray):
            if isinstance(x1, AbstractFunctionArray):
                inputs_1 = x1.inputs
                outputs_1 = x1.outputs
            else:
                return NotImplemented
        else:
            inputs_1 = None
            outputs_1 = x1

        if isinstance(x2, na.AbstractArray):
            if isinstance(x2, AbstractFunctionArray):
                inputs_2 = x2.inputs
                outputs_2 = x2.outputs
            else:
                return NotImplemented
        else:
            inputs_2 = None
            outputs_2 = x2

        if inputs_1 is not None:
            inputs = inputs_1
            if inputs_2 is not None:
                if np.any(inputs_1 != inputs_2):
                    raise InputValueError("`x1.inputs` and `x2.inputs` must be equal")
            else:
                pass
        else:
            if inputs_2 is not None:
                inputs = inputs_2
            else:
                return NotImplemented

        if out is not None:
            if not np.all(inputs == out.inputs):
                raise InputValueError("`out.inputs` must be equal to `x1.inputs` and `x2.inputs`")

        result = self.type_explicit(
            inputs=inputs,
            outputs=np.matmul(
                outputs_1,
                outputs_2,
                out=out.outputs if out is not None else out,
                **kwargs,
            )
        )

        if out is not None:
            result = out

        return result

    def __array_ufunc__(
            self,
            function: np.ufunc,
            method: str,
            *inputs,
            **kwargs,
    ) -> None | na.AbstractArray | tuple[na.AbstractArray, ...]:

        result = super().__array_ufunc__(function, method, *inputs, **kwargs)
        if result is not NotImplemented:
            return result

        nout = function.nout

        inputs_inputs = []
        inputs_outputs = []
        for inp in inputs:
            if isinstance(inp, na.AbstractArray):
                if isinstance(inp, AbstractFunctionArray):
                    inputs_inputs.append(inp.inputs)
                    inputs_outputs.append(inp.outputs)
                else:
                    if not inp.shape:
                        inputs_outputs.append(inp)
                    else:
                        return NotImplemented
            else:
                inputs_outputs.append(inp)

        for inputs_input in inputs_inputs[1:]:
            if np.any(inputs_input != inputs_inputs[0]):
                raise InputValueError(f"all inputs to {function} must have the same coordinates")

        if "out" in kwargs:
            out = kwargs.pop("out")
            outputs_out = list()
            for o in out:
                if o is not None:
                    if isinstance(o, self.type_explicit):
                        outputs_o = o.outputs
                    else:
                        raise ValueError(
                            f"each element of `out` must be an instance of {self.type_explicit}, got {type(o)}"
                        )
                else:
                    outputs_o = None
                outputs_out.append(outputs_o)
            if nout == 1:
                outputs_out = outputs_out[0]
            else:
                outputs_out = tuple(outputs_out)

            kwargs["out"] = outputs_out
        else:
            out = (None,) * nout

        if "where" in kwargs:
            where = kwargs.pop("where")
            if isinstance(where, na.AbstractArray):
                if isinstance(where, AbstractFunctionArray):
                    if np.any(where.inputs != inputs_inputs[0]):
                        raise InputValueError(f"`where.inputs` must match the rest of the inputs")
                    kwargs["where"] = where.outputs
                else:
                    return NotImplemented
            else:
                kwargs["where"] = where

        inputs_result = inputs_inputs[0]
        outputs_result = getattr(function, method)(*inputs_outputs, **kwargs)

        if nout == 1:
            outputs_result = (outputs_result,)

        result = list(
            self.type_explicit(inputs=inputs_result, outputs=outputs_result[i])
            for i in range(nout)
        )

        for i in range(nout):
            if out[i] is not None:
                out[i].inputs = result[i].inputs
                out[i].outputs = result[i].outputs
                result[i] = out[i]

        if nout == 1:
            result = result[0]
        else:
            result = tuple(result)

        return result

    def __array_function__(
            self,
            func: Callable,
            types: Collection,
            args: tuple,
            kwargs: dict[str, Any],
    ):
        result = super().__array_function__(func=func, types=types, args=args, kwargs=kwargs)
        if result is not NotImplemented:
            return result

        from . import function_array_functions

        if func in function_array_functions.DEFAULT_FUNCTIONS:
            return function_array_functions.array_function_default(func, *args, **kwargs)

        if func in function_array_functions.PERCENTILE_LIKE_FUNCTIONS:
            return function_array_functions.array_function_percentile_like(func, *args, **kwargs)

        if func in function_array_functions.ARG_REDUCE_FUNCTIONS:
            return function_array_functions.array_function_arg_reduce(func, *args, **kwargs)

        if func in function_array_functions.STACK_LIKE_FUNCTIONS:
            return function_array_functions.array_function_stack_like(func, *args, **kwargs)

        if func in function_array_functions.HANDLED_FUNCTIONS:
            return function_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    def pcolormesh(
            self,
            axs: np.ndarray,
            input_component_x: str,
            input_component_y: str,
            input_component_row: None | str = None,
            input_component_column: None | str = None,
            index: None | dict[str, int] = None,
            output_component_color: None | str = None,
            **kwargs: Any,
    ):
        """
        Plot a :class:`FunctionArray` via :func:`matplotlib.pyplot.pcolormesh`.

        :func:`FunctionArray.pcolormesh` takes in an axes object, or array of axes objects, along with components to
        be plotted along the x and y plot axes (:attr:`input_component_x` and :attr:`input_component_y`). Additional
        components can be tiled along subplot row/column and are specified in :attr:`input_component_row` and
        :attr:`input_component_column`.

        .. jupyter-execute::

            import named_arrays as na
            import numpy as np
            import astropy.units as u
            import matplotlib.pyplot as plt

            position = na.Cartesian2dVectorLinearSpace(
                start=-10,
                stop=10,
                axis=na.Cartesian2dVectorArray(
                    x='position_x',
                    y='position_y',
                ),
                num=21,
            ) * u.m

            x_width = 5 * u.m
            y_width = 2 * u.m
            velocity = 1 * u.m/u.s
            time = na.ScalarLinearSpace(
                start=0 * u.s,
                stop=3 * u.s,
                num=4,
                axis='time'
            )

            intensity = np.exp(-(((position.x + velocity*time)/x_width) ** 2 + ((position.y + 2*velocity*time)/y_width)** 2))
            scene = na.FunctionArray(
                inputs=position,
                outputs=intensity * u.DN,
            )

            fig, axs = plt.subplots(
                nrows=scene.outputs.shape['time'],
                squeeze=False,
                sharex=True,
                subplot_kw=dict(aspect='equal'),
            )
            scene.pcolormesh(
                axs=axs,
                input_component_x='x',
                input_component_y='y',
                input_component_row='time',
            )
        """



        if axs.ndim == 1:
            if input_component_row is not None:
                axs = axs[..., np.newaxis]
            if input_component_column is not None:
                axs = axs[np.newaxis, ...]

        axs = na.ScalarArray(
            ndarray=axs,
            axes=('row', 'column')
        )

        if index is None:
            index = dict()

        with astropy.visualization.quantity_support():
            for index_subplot in axs.ndindex():

                index_final = index.copy()
                if input_component_row is not None:
                    index_final[input_component_row] = index_subplot['row']
                if input_component_column is not None:
                    index_final[input_component_column] = index_subplot['column']

                inp = self[index_final].inputs

                inp_x = inp.components[input_component_x].ndarray
                inp_y = inp.components[input_component_y].ndarray

                out = self[index_final].outputs
                if output_component_color is not None:
                    out = out.components[output_component_color]

                ax = axs[index_subplot].ndarray
                ax.pcolormesh(
                    inp_x,
                    inp_y,
                    out.ndarray,
                    shading='nearest',
                    **kwargs,
                )

                if index_subplot['row'] == axs.shape['row'] - 1:
                    if isinstance(inp_x, u.Quantity):
                        ax.set_xlabel(f'{input_component_x} ({inp_x.unit})')
                    else:
                        ax.set_xlabel(f'{input_component_x}')
                else:
                    ax.set_xlabel(None)

                if index_subplot['column'] == 0:
                    if isinstance(inp_y, u.Quantity):
                        ax.set_ylabel(f'{input_component_y} ({inp_y.unit})')
                    else:
                        ax.set_ylabel(f'{input_component_y}')
                else:
                    ax.set_ylabel(None)

                if input_component_column is not None:
                    if input_component_column in inp.components:
                        if index_subplot['row'] == 0:
                            inp_column = inp.components[input_component_column]
                            ax.text(
                                x=0.5,
                                y=1.01,
                                s=f'{inp_column.mean().array.value:0.03f} {inp_column.unit:latex_inline}',
                                transform=ax.transAxes,
                                ha='center',
                                va='bottom'
                            )

                if input_component_row is not None:
                    if input_component_row in inp.components:
                        if index_subplot['column'] == axs.shape['column'] - 1:
                            inp_row = inp.components[input_component_row]
                            ax.text(
                                x=1.01,
                                y=0.5,
                                s=f'{inp_row.mean().array.value:0.03f} {inp_row.unit:latex_inline}',
                                transform=ax.transAxes,
                                va='center',
                                ha='left',
                                rotation=-90,
                            )


@dataclasses.dataclass(eq=False, repr=False)
class FunctionArray(
    AbstractFunctionArray,
    na.AbstractExplicitArray,
    Generic[InputsT, OutputsT],
):
    inputs: InputsT = dataclasses.MISSING
    outputs: OutputsT = dataclasses.MISSING

    @property
    def axes(self) -> tuple[str, ...]:
        return tuple(self.shape.keys())

    @property
    def shape(self) -> dict[str, int]:
        return na.shape_broadcasted(self.inputs, self.outputs)

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def size(self) -> int:
        return int(np.array(tuple(self.shape.values())).prod())

    @property
    def explicit(self) -> FunctionArray:
        return self.type_explicit(
            inputs=self.inputs.explicit,
            outputs=self.outputs.explicit,
        )

    def __setitem__(
            self,
            item: dict[str, int | slice | na.AbstractScalar | na.AbstractFunctionArray] | na.AbstractFunctionArray,
            value: float | u.Quantity | na.FunctionArray,
    ):

        if isinstance(item, na.AbstractFunctionArray):
            if np.any(item.inputs != self.inputs):
                raise ValueError("boolean advanced index does not have the same inputs as the array")

            item_inputs = item.outputs
            item_outputs = item.outputs

        elif isinstance(item, dict):

            item_inputs = dict()
            item_outputs = dict()
            for ax in item:
                item_ax = item[ax]
                if isinstance(item_ax, na.AbstractFunctionArray):
                    item_inputs[ax] = item_ax.inputs
                    item_outputs[ax] = item_ax.outputs
                else:
                    item_inputs[ax] = item_outputs[ax] = item_ax
        else:
            raise TypeError(
                f"`item` must be an instance of `{dict.__name__}`, or `{na.AbstractFunctionArray.__name__}`, "
                f"got `{type(item)}`"
            )

        if isinstance(value, na.AbstractArray):
            if isinstance(value, na.AbstractFunctionArray):
                value_inputs = value.inputs
                value_outputs = value.outputs
            else:
                if value.shape:
                    raise ValueError(
                        f"if `value` is an instance of `{na.AbstractArray.__name__}`, "
                        f"but not an instance of `{na.AbstractFunctionArray.__name__}`, "
                        f"`value.shape` should be empty, got {value.shape}"
                    )
                else:
                    value_inputs = None
                    value_outputs = value
        else:
            value_inputs = None
            value_outputs = value

        if value_inputs is not None:
            self.inputs[item_inputs] = value_inputs
        self.outputs[item_outputs] = value_outputs



from typing import Callable, TypeVar
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import matplotlib.artist
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
from . import scalars

__all__ = [
    "RANDOM_FUNCTIONS",
    "PLT_PLOT_LIKE_FUNCTIONS",
    "HANDLED_FUNCTIONS",
    "random",
]

RANDOM_FUNCTIONS = (
    na.random.uniform,
    na.random.normal,
    na.random.poisson,
)
PLT_PLOT_LIKE_FUNCTIONS = (
    na.plt.plot,
)
HANDLED_FUNCTIONS = dict()


def _implements(function: Callable):
    """Register a __named_array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[function] = func
        return func
    return decorator


@_implements(na.arange)
def arange(
        start: float | complex | u.Quantity | na.AbstractArray,
        stop: float | complex | u.Quantity | na.AbstractArray,
        axis: str | na.AbstractArray,
        step: int | na.AbstractArray = 1,
) -> na.ScalarArray:

    start = scalars._normalize(start)
    stop = scalars._normalize(stop)

    if start.size > 1:
        raise ValueError(f"`start` must have only one element, got shape {start.shape}")

    if stop.size > 1:
        raise ValueError(f"`stop` must have only one element, got shape {stop.shape}")

    return na.ScalarArray(
        ndarray=np.arange(
            start=start.ndarray,
            stop=stop.ndarray,
            step=step,
        ),
        axes=(axis,),
    )


def random(
        func: Callable,
        *args: float | u.Quantity | na.AbstractScalarArray,
        shape_random: None | dict[str, int] = None,
        seed: None | int = None,
        **kwargs: float | u.Quantity | na.AbstractScalarArray,
) -> na.ScalarArray:

    try:
        args = tuple(scalars._normalize(arg) for arg in args)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:
        return NotImplemented

    if shape_random is None:
        shape_random = dict()

    shape_base = na.shape_broadcasted(*args, *kwargs.values())
    shape = na.broadcast_shapes(shape_base, shape_random)

    args = tuple(arg.ndarray_aligned(shape) for arg in args)
    kwargs = {k: kwargs[k].ndarray_aligned(shape) for k in kwargs}

    unit = None
    for a in args + tuple(kwargs.values()):
        if isinstance(a, u.Quantity):
            unit = a.unit
            break

    if unit is not None:
        args = tuple(
            arg.value if isinstance(arg, u.Quantity)
            else (arg << u.dimensionless_unscaled).to_value(unit)
            for arg in args
        )
        kwargs = {
            k: kwargs[k].value if isinstance(kwargs[k], u.Quantity)
            else (kwargs[k] << u.dimensionless_unscaled).to_value(unit)
            for k in kwargs
        }

    if seed is None:
        func = getattr(np.random, func.__name__)
    else:
        func = getattr(np.random.default_rng(seed), func.__name__)

    value = func(
        *args,
        size=tuple(shape.values()),
        **kwargs,
    )

    if unit is not None:
        value = value << unit

    return na.ScalarArray(
        ndarray=value,
        axes=tuple(shape.keys()),
    )


def plt_plot_like(
        func: Callable,
        *args: na.AbstractScalarArray,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalarArray = True,
        **kwargs,
) -> na.ScalarArray[npt.NDArray[None | matplotlib.artist.Artist]]:

    try:
        args = tuple(scalars._normalize(arg) for arg in args)
        where = scalars._normalize(where)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:
        return NotImplemented

    if ax is None:
        ax = plt.gca()
    ax = na.as_named_array(ax)

    shape = na.shape_broadcasted(*args)

    if axis is None:
        if len(shape) != 1:
            raise ValueError(
                f"if `axis` is `None`, the broadcasted shape of `*args`, {shape}, should have one element"
            )
        axis = next(iter(shape))

    shape_orthogonal = {a: shape[a] for a in shape if a != axis}

    args = tuple(arg.broadcast_to(shape) for arg in args)

    if not set(ax.shape).issubset(shape_orthogonal):
        raise ValueError(
            f"the shape of `ax`, {ax.shape}, "
            f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}",
        )
    ax = ax.broadcast_to(shape_orthogonal)

    if not set(where.shape).issubset(shape_orthogonal):
        raise ValueError(
            f"the shape of `where`, {where.shape}, "
            f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}"
        )
    where = where.broadcast_to(shape_orthogonal)

    kwargs_broadcasted = dict()
    for k in kwargs:
        kwarg = kwargs[k]
        if not set(na.shape(kwarg)).issubset(shape_orthogonal):
            raise ValueError(
                f"the shape of `{k}`, {na.shape(kwarg)}, "
                f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}"
            )
        kwargs_broadcasted[k] = na.broadcast_to(kwarg, shape_orthogonal)
    kwargs = kwargs_broadcasted

    result = na.ScalarArray.empty(shape=shape_orthogonal, dtype=object)

    for index in na.ndindex(shape_orthogonal):
        if where[index]:
            func_matplotlib = getattr(ax[index].ndarray, func.__name__)
            args_index = tuple(arg[index].ndarray for arg in args)
            kwargs_index = {k: kwargs[k][index].ndarray for k in kwargs}
            result[index] = func_matplotlib(
                *args_index,
                **kwargs_index,
            )[0]

    return result



from __future__ import annotations
from typing import Type
from typing_extensions import Self
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractSpectralPositionVectorArray',
    'SpectralPositionVectorArray',
    'AbstractImplicitSpectralPositionVectorArray',
    'AbstractParameterizedSpectralPositionVectorArray',
    'AbstractSpectralPositionVectorSpace',
    'SpectralPositionVectorLinearSpace',
]

@dataclasses.dataclass(eq=False, repr=False)
class AbstractSpectralPositionVectorArray(
    na.AbstractPositionVectorArray,
    na.AbstractSpectralVectorArray,
):

    @property
    def type_abstract(self) -> Type[na.AbstractArray]:
        return AbstractSpectralPositionVectorArray

    @property
    def type_explicit(self) -> Type[na.AbstractExplicitArray]:
        return SpectralPositionVectorArray

    @property
    def type_matrix(self) -> Type[na.SpectralPositionMatrixArray]:
        return na.SpectralPositionMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class SpectralPositionVectorArray(
    AbstractSpectralPositionVectorArray,
    na.PositionVectorArray,
    na.SpectralVectorArray,
):

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> SpectralPositionVectorArray:
        result = super().from_scalar(scalar, like=like)
        if result is not NotImplemented:
            return result
        return cls(wavelength=scalar, position=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitSpectralPositionVectorArray(
    AbstractSpectralPositionVectorArray,
    na.AbstractImplicitPositionVectorArray,
    na.AbstractImplicitSpectralVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedSpectralPositionVectorArray(
    AbstractImplicitSpectralPositionVectorArray,
    na.AbstractParameterizedVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSpectralPositionVectorSpace(
    AbstractParameterizedSpectralPositionVectorArray,
    na.AbstractVectorSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class SpectralPositionVectorLinearSpace(
    AbstractSpectralPositionVectorSpace,
    na.AbstractVectorLinearSpace,
):
    pass


from __future__ import annotations
from typing import Type, Generic, TypeVar
from typing_extensions import Self
import abc
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractPositionVectorArray',
    'PositionVectorArray',
    'AbstractImplicitPositionVectorArray',
    'AbstractParameterizedPositionVectorArray',
    'AbstractPositionVectorSpace',
    'PositionVectorLinearSpace',

]

PositionT = TypeVar("PositionT", bound=na.ArrayLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPositionVectorArray(
    na.AbstractCartesianVectorArray,
):

    @property
    @abc.abstractmethod
    def position(self) -> na.ArrayLike:
        """
        The `position` component of the vector.
        """

    @property
    def type_abstract(self) -> Type[na.AbstractArray]:
        return AbstractPositionVectorArray

    @property
    def type_explicit(self) -> Type[na.AbstractExplicitArray]:
        return PositionVectorArray

    @property
    def type_matrix(self) -> Type[na.PositionMatrixArray]:
        return na.PositionMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class PositionVectorArray(
    AbstractPositionVectorArray,
    na.AbstractExplicitCartesianVectorArray,
    Generic[PositionT],
):
    position: PositionT = 0

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> PositionVectorArray:
        result = super().from_scalar(scalar, like=like)
        if result is not NotImplemented:
            return result
        return cls(Position=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitPositionVectorArray(
    AbstractPositionVectorArray,
    na.AbstractImplicitCartesianVectorArray,
):

    @property
    def position(self) -> na.ArrayLike:
        return self.explicit.position


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedPositionVectorArray(
    AbstractImplicitPositionVectorArray,
    na.AbstractParameterizedVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPositionVectorSpace(
    AbstractParameterizedPositionVectorArray,
    na.AbstractVectorSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class PositionVectorLinearSpace(
    AbstractPositionVectorSpace,
    na.AbstractVectorLinearSpace,
):
    pass



from __future__ import annotations
from typing import Type, Sequence
import dataclasses
import named_arrays as na

__all__ = [
    "AbstractCartesianNdVectorArray",
    "CartesianNdVectorArray"
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesianNdVectorArray(
    na.AbstractCartesianVectorArray,
):

    @property
    def type_abstract(self) -> Type[AbstractCartesianNdVectorArray]:
        return AbstractCartesianNdVectorArray

    @property
    def type_explicit(self) -> Type[CartesianNdVectorArray]:
        return CartesianNdVectorArray

    @property
    def type_matrix(self) -> Type[na.CartesianNdMatrixArray]:
        return na.CartesianNdMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class CartesianNdVectorArray(
    AbstractCartesianNdVectorArray,
    na.AbstractExplicitCartesianVectorArray,
):
    components: dict[str, na.ArrayLike]

    @classmethod
    def from_scalar(
            cls,
            scalar: na.ScalarLike,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> CartesianNdVectorArray:

        if like is None:
            raise ValueError("like argument must be specified for CartesianNdArrays")

        result = super().from_scalar(scalar, like=like)
        if result is not NotImplemented:
            return result
        else:
            raise ValueError("all implementations of from_scalar return NotImplemented")



    @classmethod
    def from_components(cls, components: dict[str, na.ArrayLike]) -> na.CartesianNdMatrixArray:
        return cls(components)



from __future__ import annotations
from typing import TypeVar, Generic, Type
import abc
import dataclasses
import named_arrays as na
import numpy as np
import astropy.units as u

__all__ = [
    'AbstractSpectralPositionMatrixArray',
    'SpectralPositionMatrixArray',
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSpectralPositionMatrixArray(
    na.AbstractSpectralPositionVectorArray,
):

    @property
    def type_abstract(self) -> Type[AbstractSpectralPositionMatrixArray]:
        return AbstractSpectralPositionMatrixArray

    @property
    def type_explicit(self) -> Type[SpectralPositionMatrixArray]:
        return SpectralPositionMatrixArray

    @property
    def type_vector(self) -> Type[na.SpectralPositionVectorArray]:
        return na.SpectralPositionVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        return NotImplementedError

    @property
    def inverse(self) -> na.AbstractMatrixArray:
        return NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class SpectralPositionMatrixArray(
    na.SpectralPositionVectorArray,
    AbstractSpectralPositionMatrixArray,
    na.AbstractExplicitMatrixArray,
):
    pass


from typing import Callable
import numpy.typing as npt
import matplotlib.axes
import matplotlib.artist
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
import named_arrays._scalars.scalar_named_array_functions
from . import uncertainties

__all__ = [
    "RANDOM_FUNCTIONS",
    "PLT_PLOT_LIKE_FUNCTIONS",
    "HANDLED_FUNCTIONS",
    "random",
]

RANDOM_FUNCTIONS = named_arrays._scalars.scalar_named_array_functions.RANDOM_FUNCTIONS
PLT_PLOT_LIKE_FUNCTIONS = named_arrays._scalars.scalar_named_array_functions.PLT_PLOT_LIKE_FUNCTIONS
HANDLED_FUNCTIONS = dict()


def _implements(function: Callable):
    """Register a __named_array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[function] = func
        return func
    return decorator


def random(
        func: Callable,
        *args: float | u.Quantity | na.AbstractScalar,
        shape_random: None | dict[str, int] = None,
        seed: None | int = None,
        **kwargs: float | u.Quantity | na.AbstractScalar,
) -> na.UncertainScalarArray:

    try:
        args = tuple(uncertainties._normalize(arg) for arg in args)
        kwargs = {k: uncertainties._normalize(kwargs[k]) for k in kwargs}
    except na.UncertainScalarTypeError:
        return NotImplemented

    return na.UncertainScalarArray(
        nominal=func(
            *tuple(arg.nominal for arg in args),
            shape_random=shape_random,
            seed=seed,
            **{k: kwargs[k].nominal for k in kwargs},
        ),
        distribution=func(
            *tuple(arg.distribution for arg in args),
            shape_random=shape_random,
            seed=seed,
            **{k: kwargs[k].distribution for k in kwargs},
        )
    )


def plt_plot_like(
        func: Callable,
        *args: na.AbstractScalar,
        ax: None | matplotlib.axes.Axes = None,
        axis: None | str = None,
        where: bool | na.AbstractScalarArray = True,
        **kwargs,
) -> na.UncertainScalarArray[
    npt.NDArray[matplotlib.artist.Artist],
    npt.NDArray[matplotlib.artist.Artist]
]:
    try:
        args = tuple(uncertainties._normalize(arg) for arg in args)
        where = uncertainties._normalize(where)
        kwargs = {k: uncertainties._normalize(kwargs[k]) for k in kwargs}
    except na.UncertainScalarTypeError:
        return NotImplemented

    shape = na.shape_broadcasted(*args)

    if axis is None:
        if len(shape) != 1:
            raise ValueError(
                f"if `axis` is `None`, the broadcasted shape of `*args`, {shape}, should have one element"
            )
        axis = next(iter(shape))

    args = tuple(arg.broadcast_to(shape) for arg in args)

    alpha = max(1 / args[~0].num_distribution, 1/255)
    if "alpha" in kwargs:
        kwargs["alpha"] *= alpha
    else:
        kwargs["alpha"] = na.UncertainScalarArray(1, alpha)

    if "color" not in kwargs:
        color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
        shape_orthogonal = {a: shape[a] for a in shape if a != axis}
        color = na.ScalarArray.empty(shape=shape_orthogonal, dtype=object)
        for i, index in enumerate(color.ndindex()):
            color[index] = color_cycle[i % len(color_cycle)]
        kwargs["color"] = uncertainties._normalize(color)

    result = na.UncertainScalarArray(
        nominal=func(
            *tuple(na.as_named_array(arg.nominal) for arg in args),
            ax=ax,
            axis=axis,
            where=where.nominal,
            **{k: kwargs[k].nominal for k in kwargs}
        ),
        distribution=func(
            *tuple(na.as_named_array(arg.distribution) for arg in args),
            ax=ax,
            axis=axis,
            where=where.distribution,
            **{k: kwargs[k].distribution for k in kwargs}
        )
    )

    return result



from __future__ import annotations
from typing import Sequence, overload, Type, Any, Callable, TypeVar
import functools
import numpy as np
import astropy.units as u
import named_arrays as na

__all__ = [
    '_named_array_function',
    'arange',
    'linspace',
    'logspace',
    'geomspace',
    'ndim',
    'shape',
    'broadcast_to',
    'stack',
    'concatenate',
    'add_axes',
]

AxisT = TypeVar("AxisT", bound="str | na.AbstractArray")
NumT = TypeVar("NumT", bound="int | na.AbstractArray")
BaseT = TypeVar("BaseT", bound="int | na.AbstractArray")


def _is_subclass(a: Any, b: Any):
    if type(a) == type(b):
        return 0
    elif isinstance(a, type(b)):
        return 1
    elif isinstance(b, type(a)):
        return -1
    else:
        return 0


def _named_array_function(func: Callable, *args, **kwargs):
    arrays_args = tuple(arg for arg in args if isinstance(arg, na.AbstractArray))
    arrays_kwargs = tuple(kwargs[k] for k in kwargs if isinstance(kwargs[k], na.AbstractArray))
    arrays = arrays_args + arrays_kwargs

    arrays = sorted(arrays, key=functools.cmp_to_key(_is_subclass))

    for array in arrays:
        res = array.__named_array_function__(func, *args, **kwargs)
        if res is not NotImplemented:
            return res

    raise TypeError("all types returned `NotImplemented`")


def arange(
        start: float | complex | u.Quantity | na.AbstractArray,
        stop: float | complex | u.Quantity | na.AbstractArray,
        axis: str | na.AbstractArray,
        step: int | na.AbstractArray = 1,
) -> na.AbstractExplicitArray:
    """
    Redefined version of :func:`numpy.arange` with an ``axis`` parameter.

    Parameters
    ----------
    start
        starting value of the sequence
    stop
        ending value of the sequence
    axis
        name of the new sequence axis
    step
        step size between consecutive elements of the sequence

    See Also
    --------
    :func:`numpy.arange` : Equivalent numpy function

    :class:`named_arrays.ScalarArrayRange` : Implicit array version of this function for scalars
    """
    return _named_array_function(
        func=arange,
        start=na.as_named_array(start),
        stop=na.as_named_array(stop),
        axis=axis,
        step=step,
    )


def linspace(
        start: na.StartT,
        stop: na.StopT,
        axis: AxisT,
        num: NumT = 50,
        endpoint: bool = True,
        dtype: None | type | np.dtype = None,
) -> na.StartT | na.StopT | AxisT | NumT:
    """
    Create an array of evenly-spaced numbers between :attr:`start` and :attr:`stop`

    Parameters
    ----------
    start
        The starting value of the sequence
    stop
        The last value of the sequence, unless :attr:`endpoint` is :class:`False`
    axis
        The name of the new axis corresponding to the sequence
    num
        Number of values in the sequence
    endpoint
        Flag controlling whether :attr:`stop` should be included in the sequence.
    dtype
        :class:`numpy.dtype` of the result

    See Also
    --------
    :func:`numpy.linspace` :  Corresponding numpy function.

    :class:`named_arrays.ScalarLinearSpace` : Corresponding implicit scalar array.

    :class:`named_arrays.UncertainScalarLinearSpace`: Corresponding implicit uncertain scalar array.
    """
    return np.linspace(
        start=na.as_named_array(start),
        stop=stop,
        axis=axis,
        num=num,
        endpoint=endpoint,
        dtype=dtype,
    )


def logspace(
        start: na.StartT,
        stop: na.StopT,
        axis: AxisT,
        num: NumT = 50,
        endpoint: bool = True,
        base: BaseT = 10,
        dtype: None | type | np.dtype = None,
) -> na.StartT | na.StopT | AxisT | NumT:
    """
    Create an array of evenly-spaced numbers on a log scale between :attr:`start` and :attr:`stop`

    Parameters
    ----------
    start
        The starting value of the sequence
    stop
        The last value of the sequence, unless :attr:`endpoint` is :class:`False`
    axis
        The name of the new axis corresponding to the sequence
    num
        Number of values in the sequence
    endpoint
        Flag controlling whether :attr:`stop` should be included in the sequence.
    base
        The base of the logarithm
    dtype
        :class:`numpy.dtype` of the result

    See Also
    --------
    :func:`numpy.logspace` :  Corresponding numpy function.

    :class:`named_arrays.ScalarLogarithmicSpace` : Corresponding implicit scalar array.

    :class:`named_arrays.UncertainScalarLogarithmicSpace`: Corresponding implicit uncertain scalar array.
    """
    return np.logspace(
        start=na.as_named_array(start),
        stop=stop,
        axis=axis,
        num=num,
        endpoint=endpoint,
        base=base,
        dtype=dtype,
    )


def geomspace(
        start: na.StartT,
        stop: na.StopT,
        axis: AxisT,
        num: NumT = 50,
        endpoint: bool = True,
        dtype: None | type | np.dtype = None,
) -> na.StartT | na.StopT | AxisT | NumT:
    """
    Create an array of a geometric progression of numbers between :attr:`start` and :attr:`stop`

    Parameters
    ----------
    start
        The starting value of the sequence
    stop
        The last value of the sequence, unless :attr:`endpoint` is :class:`False`
    axis
        The name of the new axis corresponding to the sequence
    num
        Number of values in the sequence
    endpoint
        Flag controlling whether :attr:`stop` should be included in the sequence.
    dtype
        :class:`numpy.dtype` of the result

    See Also
    --------
    :func:`numpy.geomspace` :  Corresponding numpy function.

    :class:`named_arrays.ScalarGeometricSpace` : Corresponding implicit scalar array.

    :class:`named_arrays.UncertainScalarGeometricSpace`: Corresponding implicit uncertain scalar array.
    """
    return np.geomspace(
        start=na.as_named_array(start),
        stop=stop,
        axis=axis,
        num=num,
        endpoint=endpoint,
        dtype=dtype,
    )


def ndim(a: na.AbstractArray) -> int:
    return np.ndim(a)


def shape(a: na.ArrayLike) -> dict[str, int]:
    if not isinstance(a, na.AbstractArray):
        a = na.ScalarArray(a)
    return np.shape(a)


@overload
def broadcast_to(array: float | complex | np.ndarray | u.Quantity, shape: dict[str, int]) -> na.ScalarArray:
    ...


@overload
def broadcast_to(array: na.AbstractScalarArray, shape: dict[str, int]) -> na.ScalarArray:
    ...


def broadcast_to(array, shape):
    if not isinstance(array, na.AbstractArray):
        array = na.ScalarArray(array)
    return np.broadcast_to(array=array, shape=shape)


def stack(
        arrays: Sequence[na.ArrayLike],
        axis: str,
        out: None | na.AbstractExplicitArray = None,
        *,
        dtype: str | np.dtype | Type = None,
        casting: None | str = "same_kind",
) -> na.AbstractArray:
    if not any(isinstance(a, na.AbstractArray) for a in arrays):
        arrays = list(arrays)
        arrays[0] = na.ScalarArray(arrays[0])
    return np.stack(
        arrays=arrays,
        axis=axis,
        out=out,
        dtype=dtype,
        casting=casting,
    )


def concatenate(
        arrays: Sequence[na.ArrayLike],
        axis: str,
        out: None | na.AbstractExplicitArray = None,
        *,
        dtype: str | np.dtype | Type = None,
        casting: None | str = "same_kind",
) -> na.AbstractArray:
    if not any(isinstance(a, na.AbstractArray) for a in arrays):
        arrays = list(arrays)
        arrays[0] = na.ScalarArray(arrays[0])
    return np.concatenate(
        arrays=arrays,
        axis=axis,
        out=out,
        dtype=dtype,
        casting=casting,
    )


def add_axes(array: na.ArrayLike, axes: str | Sequence[str]):
    if not isinstance(array, na.AbstractArray):
        array = na.ScalarArray(array)
    return array.add_axes(axes)



from __future__ import annotations
from typing import TypeVar, Generic, Type
import abc
import dataclasses
import named_arrays as na
import numpy as np

__all__ = [
    'AbstractPositionMatrixArray',
    'PositionMatrixArray',
]

PositionT = TypeVar('PositionT', bound=na.AbstractVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPositionMatrixArray(
    na.AbstractCartesianMatrixArray,
    na.AbstractPositionVectorArray,
):
    @property
    @abc.abstractmethod
    def position(self) -> na.AbstractVectorArray:
        """
        The `position` component of the matrix.
        """

    @property
    def type_abstract(self) -> Type[AbstractPositionMatrixArray]:
        return AbstractPositionMatrixArray

    @property
    def type_explicit(self) -> Type[PositionMatrixArray]:
        return PositionMatrixArray

    @property
    def type_vector(self) -> Type[na.PositionVectorArray]:
        return na.PositionVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        raise NotImplementedError

    @property
    def inverse(self) -> na.AbstractMatrixArray:
        raise NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class PositionMatrixArray(
    na.PositionVectorArray,
    AbstractPositionMatrixArray,
    na.AbstractExplicitMatrixArray,
    Generic[PositionT],
):
    pass



from __future__ import annotations
from typing import TypeVar, Generic, Type
import abc
import dataclasses
import named_arrays as na

__all__ = [
    "AbstractCartesianNdMatrixArray",
    "CartesianNdMatrixArray",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesianNdMatrixArray(
    na.AbstractCartesianMatrixArray,
    na.AbstractCartesianNdVectorArray,
):

    @property
    def type_abstract(self) -> Type[AbstractCartesianNdMatrixArray]:
        return AbstractCartesianNdMatrixArray

    @property
    def type_explicit(self) -> Type[CartesianNdMatrixArray]:
        return CartesianNdMatrixArray

    @property
    def type_vector(self) -> Type[na.CartesianNdVectorArray]:
        return na.CartesianNdVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        return NotImplementedError

    @property
    def inverse(self) -> na.AbstractMatrixArray:
        return NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class CartesianNdMatrixArray(
    na.CartesianNdVectorArray,
    AbstractCartesianNdMatrixArray,
    na.AbstractExplicitCartesianMatrixArray,
):
    pass
